import os
import json
import textwrap
from typing import List, Dict, Any, Optional, Tuple, Union
from metaflow import Step
from metaflow_extensions.netflix_ext.cmd.debug.constants import Constants
from metaflow_extensions.netflix_ext.cmd.debug.debug_utils import fetch_environment_type
from metaflow_extensions.netflix_ext.cmd.debug.debug_stub_generator import (
    DebugStubGenerator,
)


class DebugScriptGenerator(object):
    def __init__(self, task_pathspec: str, inspect: Union[str, bool]):
        """
        Initializes the debugScriptGenerator.

        Parameters
        ----------
        task_pathspec : str
            The pathspec of the task.
        inspect : Union[str, bool]
            The inspect flag.
        """
        self.task_pathspec = task_pathspec
        if isinstance(inspect, bool):
            self.inspect = inspect
        else:
            self.inspect = True if inspect.lower() in ("true", "1") else False
        self.debug_stub_generator = DebugStubGenerator(task_pathspec)

    def namespace_imports(self, script: str) -> str:
        """
        Appends the namespace imports to the script.

        Parameters
        ----------
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the namespace imports appended.
        """
        script += textwrap.dedent(
            Constants.CURRENT_TASK_NAMESPACE_STUB.format(
                Namespace=self.debug_stub_generator.task_namespace
            )
        )
        return script

    def append_imports(self, metaflow_root_dir: str, script: str) -> str:
        """
        Appends the imports to the script.

        Parameters
        ----------
        metaflow_root_dir : str
            The root directory of the Metaflow installation.
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the imports appended.
        """
        script += textwrap.dedent(
            Constants.IMPORTS_STUB.format(
                MetaflowEnvEscapeDir=os.path.join(
                    metaflow_root_dir, "_escape_trampolines"
                ),
                MetaflowFileName=self.debug_stub_generator.file_name.split(".")[0],
            )
        )
        return script

    def append_debug_stub_instantiation(self, script: str) -> str:
        """
        Appends the debug stub generator instantiation to the script.

        Parameters
        ----------
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the debug stub generator instantiation appended.
        """
        script += textwrap.dedent(
            Constants.DEBUG_STUB_GENERATOR_INSTANTIATION_TEMPLATE.format(
                TaskPathspec=self.task_pathspec
            )
        )
        return script

    @staticmethod
    def append_stubbed_classes(flow_name: str, script: str) -> str:
        """
        Appends the stubs to the script.

        Parameters
        ----------
        flow_name : str
            The name of the flow.
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the stubs appended.
        """
        script += textwrap.dedent(
            Constants.DEBUG_INPUT_ITEM_STUB.format(
                FlowSpecClass=flow_name,
            )
        )
        script += textwrap.dedent(
            Constants.DEBUG_INPUT_STUB.format(
                FlowSpecClass=flow_name,
            )
        )
        script += textwrap.dedent(
            Constants.DEBUG_LINEAR_STEP_STUB.format(
                FlowSpecClass=flow_name,
            )
        )
        script += textwrap.dedent(
            Constants.DEBUG_JOIN_STEP_STUB.format(
                FlowSpecClass=flow_name,
            )
        )
        return script

    def append_join_step_instantiation(self, script: str) -> str:
        """
        Appends the join step instantiation to the script.

        Parameters
        ----------
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the join step instantiation appended.
        """
        template = (
            Constants.JOIN_STUB_INSPECT_INSTANTIATION_TEMPLATE
            if self.inspect
            else Constants.JOIN_STUB_INSTANTIATION_TEMPLATE
        )
        script += textwrap.dedent(template)
        return script

    def append_linear_step_instantiation(self, script: str) -> str:
        """
        Appends the linear step instantiation to the script.

        Parameters
        ----------
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the linear step instantiation appended.
        """
        template = (
            Constants.LINEAR_STUB_INSPECT_INSTANTIATION_TEMPLATE
            if self.inspect
            else Constants.LINEAR_STUB_INSTANTIATION_TEMPLATE
        )
        script += textwrap.dedent(template)
        return script

    def append_start_step_instantiation(self, script: str) -> str:
        """
        Appends the start step instantiation to the script.

        Parameters
        ----------
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the start step instantiation appended.
        """
        _run_pathspec = "/".join(self.task_pathspec.split("/")[:2])
        prev_task = Step(
            os.path.join(_run_pathspec, "_parameters"), _namespace_check=False
        ).task
        template = (
            Constants.START_STUB_INSPECT_INSTANTIATION_TEMPLATE
            if self.inspect
            else Constants.START_STUB_INSTANTIATION_TEMPLATE
        )
        script += textwrap.dedent(template.format(ParameterTask=prev_task.pathspec))
        return script

    def append_current_stub_instantiation(self, script: str) -> str:
        """
        Appends the current stub instantiation to the script.

        Parameters
        ----------
        script : str
            The script for the step.

        Returns
        -------
        str
            The script with the current stub instantiation appended.
        """
        script += textwrap.dedent(
            Constants.CURRENT_STUB_INSTANTIATION_TEMPLATE.format(
                TaskPathspec=self.task_pathspec
            )
        )
        return script

    @staticmethod
    def write_debug_script(
        metaflow_root_dir: str, debug_script: str, debug_file_name: str
    ) -> None:
        """
        Writes the debug script for the step.

        Parameters
        ----------
        metaflow_root_dir : str
            The root directory of the Metaflow installation.
        debug_script : str
            The debug script for the step.
        debug_file_name : str
            The name of the debug script.
        """
        debug_script_path = os.path.join(metaflow_root_dir, debug_file_name)
        with open(debug_script_path, "w") as f:
            f.write(debug_script)

    def generate_debug_script(self, metaflow_root_dir: str, flow_name: str) -> str:
        """
        Generates the debug script for the step.

        Parameters
        ----------
        metaflow_root_dir : str
            The root directory of the Metaflow installation.
        flow_name : str
            The name of the flow.

        Returns
        -------
        str
            The debug script for the step.
        """
        print("Generating debug script for task: {}".format(self.task_pathspec))
        # We first write out the imports
        debug_script = self.append_imports(metaflow_root_dir, "")

        # We then write out the namespace imports
        debug_script = self.namespace_imports(debug_script)

        # We then write out the stubbed classes
        debug_script = self.append_stubbed_classes(flow_name, debug_script)

        # We instantiate the current stub generator
        debug_script = self.append_current_stub_instantiation(debug_script)

        # We instantiate the debug stub generator
        debug_script = self.append_debug_stub_instantiation(debug_script)

        # We then write out the stubs instantiation
        if self.debug_stub_generator.step_type == "start":
            debug_script = self.append_start_step_instantiation(debug_script)
        elif self.debug_stub_generator.step_type == "join":
            debug_script = self.append_join_step_instantiation(debug_script)
        else:
            debug_script = self.append_linear_step_instantiation(debug_script)

        return debug_script

    def generate_debug_notebook(
        self,
        metaflow_root_dir: str,
        debug_file_name: str,
        kernel_def: Optional[Tuple[str, str]] = None,
    ) -> Dict[str, Any]:
        """
        Generates the debug notebook for the task.

        Parameters
        ----------
        metaflow_root_dir : str
            The root directory of the Metaflow installation.
        debug_file_name : str
            The name of the debug script
        kernel_def : Optional[Tuple[str, str]], optional, default None
            The kernel definition containing the kernel name and the kernel display name.

        Returns
        -------
        Dict[str, Any]
            The dictionary representing the debug notebook for the task.
        """
        # We first generate the notebook cells
        notebook_cells = self.generate_notebook_cells(
            metaflow_root_dir, debug_file_name
        )

        # We then generate the notebook JSON
        notebook_json = self.generate_notebook_json(notebook_cells, kernel_def)
        return notebook_json

    @staticmethod
    def write_debug_notebook(
        metaflow_root_dir: str, notebook_json: Dict[str, Any]
    ) -> None:
        """
        Writes the debug notebook for the task.

        Parameters
        ----------
        metaflow_root_dir : str
            The root directory of the Metaflow installation.
        notebook_json : Dict[str, Any]
            The name of the debug notebook.
        """
        # We then write out the notebook JSON
        debug_notebook_name = Constants.DEBUG_SCRIPT_NAME
        debug_notebook_path = os.path.join(metaflow_root_dir, debug_notebook_name)
        with open(debug_notebook_path, "w") as f:
            json.dump(notebook_json, f, indent=4)

    @staticmethod
    def generate_notebook_json(
        notebook_cells: List[Dict[str, Any]],
        kernel_def: Optional[Tuple[str, str]] = None,
    ) -> Dict[str, Any]:
        """
        Generates the notebook JSON for the task.

        Parameters
        ----------
        notebook_cells : List[Dict[str, Any]]
            The list of notebook cells.
        kernel_def : Optional[Tuple[str, str]], optional, default None
            The kernel definition containing the kernel name and the kernel display name.

        Returns
        -------
        Dict[str, Any]
            The notebook JSON for the task.
        """
        # Create an empty notebook
        nb = {
            "cells": [],
            "metadata": (
                {}
                if kernel_def is None
                else {
                    "kernelspec": {
                        "name": kernel_def[0],
                        "display_name": kernel_def[1],
                        "language": "python",
                    }
                }
            ),
            "nbformat": 4,
            "nbformat_minor": 4,
        }

        # Add cells to the notebook
        for cell in notebook_cells:
            nb_cell = {
                "cell_type": cell.get("format"),
                "execution_count": None,
                "metadata": {},
                "outputs": [],
                "source": [cell.get("content")],
            }
            nb["cells"].append(nb_cell)

        return nb

    def get_function_definition(self, metaflow_root_dir: str) -> Tuple[str, int]:
        """
        Returns the function definition and the end line number for the task.

        Parameters
        ----------
        metaflow_root_dir : str
            The root directory of the Metaflow installation.

        Returns
        -------
        Tuple[str, int]
            The function definition and the end line number for the task.
        """
        with open(
            os.path.join(metaflow_root_dir, self.debug_stub_generator.file_name), "r"
        ) as f:
            step_started = False
            function_definition = ""
            is_step_end = self.debug_stub_generator.task.parent.id == "end"
            start_indentation = None
            for i, line in enumerate(f):
                if i == self.debug_stub_generator.task_line_num:
                    step_started = True
                    start_indentation = len(line) - len(line.lstrip())

                if step_started:
                    current_indentation = len(line) - len(line.lstrip())
                    if "self.next(self." in line:
                        function_definition += line
                        return function_definition, i

                    # Special case for the end step, as there is no next step and we need to
                    # return the function definition at the end of the file or when the function ends
                    if is_step_end:
                        if line.strip() and current_indentation < start_indentation:
                            return function_definition, i - 1

                    function_definition += line

        # If we reach here, it means that the step is the last step in the flow
        # and there is no other user code after this step
        return function_definition, i

    def generate_notebook_cells(
        self, metaflow_root_dir: str, debug_file_name: str
    ) -> List[Dict[str, Any]]:
        """
        Generates the notebook cells for the task.

        Parameters
        ----------
        metaflow_root_dir : str
            The root directory of the Metaflow installation.
        debug_file_name : str
            The name of the debug notebook.

        Returns
        -------
        str
            The notebook cells for the task.
        """
        # We first add a markdown cell stating that it is a debugging notebook
        notebook_cells = [
            {
                "format": "markdown",
                "content": Constants.JUPYTER_TITLE_MARKDOWN.format(
                    TaskPathSpec=self.task_pathspec,
                    DebugType=(
                        "Post-Execution State"
                        if self.inspect
                        else "Pre-Execution State"
                    ),
                ),
            }
        ]

        # We now add a code cell that incorporates the Metaflow escape hatch
        # We only add this if the user code ran in a conda environment
        if fetch_environment_type(self.debug_stub_generator.task) != "default":
            script = textwrap.dedent(
                Constants.ESCAPE_HATCH_STUB.format(
                    MetaflowEnvEscapeDir=os.path.abspath(
                        os.path.join(metaflow_root_dir, "_escape_trampolines")
                    )
                )
            )
            notebook_cells.append({"format": "code", "content": script.strip()})

        # We now add a code cell that imports the debugging modules
        script = "from {} import * ".format(debug_file_name.split(".")[0])
        notebook_cells.append({"format": "code", "content": script})

        # We now add a markdown cell that tells the user to add their own code
        flow_file_name = self.debug_stub_generator.file_name
        start_line_num = self.debug_stub_generator.task_line_num
        function_definition, end_line_num = self.get_function_definition(
            metaflow_root_dir
        )
        markdown_content = Constants.JUPYTER_INSTRUCTIONS_MARKDOWN.format(
            FileName=flow_file_name,
            StepStartLine=start_line_num,
            StepEndLine=end_line_num,
            FunctionDefinition=textwrap.dedent(function_definition),
        )
        notebook_cells.append(
            {"format": "markdown", "content": textwrap.dedent(markdown_content.strip())}
        )

        # Add an empty cell for the user to add their own code
        notebook_cells.append({"format": "code", "content": ""})
        return notebook_cells



import glob
import os
import re
import shutil
import subprocess
import sys

from functools import partial
from itertools import chain
from pathlib import Path

WHITELIST = {
    "README.txt",
    "vendor_any.txt",
    "pip.LICENSE",
}

# Borrowed from https://github.com/pypa/pip/tree/main/src/pip/_vendor

VENDOR_SUBDIR = re.compile(r"^vendor/vendor_([a-zA-Z0-9_]+).txt$")

MF_VENDOR_PREFIX = "metaflow._vendor"
TL_PACKAGE_NAME = "metaflow_extensions"


def delete_all(*paths, whitelist=frozenset()):
    for item in paths:
        if item.is_dir():
            shutil.rmtree(item, ignore_errors=True)
        elif item.is_file() and item.name not in whitelist:
            item.unlink()


def find_mf_vendored(*paths, exclude_dirs=frozenset()):
    to_return = []
    for item in paths:
        if item.is_dir() and item.name not in exclude_dirs:
            to_return.append(item.name)
        elif item.is_file() and item.suffix in (".py", ".LICENSE"):
            to_return.append(item.name)
    return to_return


def iter_subtree(path):
    """Recursively yield all files in a subtree, depth-first"""
    if not path.is_dir():
        if path.is_file():
            yield path
        return
    for item in path.iterdir():
        if item.is_dir():
            yield from iter_subtree(item)
        elif item.is_file():
            yield item


def patch_vendor_imports(file, replacements):
    text = file.read_text("utf8")
    for replacement in replacements:
        text = replacement(text)
    file.write_text(text, "utf8")


def find_vendored_libs(vendor_dir, whitelist, whitelist_dirs):
    vendored_libs = []
    paths = []
    for item in vendor_dir.iterdir():
        if item.is_dir() and not item in whitelist_dirs:
            vendored_libs.append(item.name)
        elif item.is_file() and item.name not in whitelist:
            vendored_libs.append(item.stem)  # without extension
        else:  # not a dir or a file not in the whilelist
            continue
        paths.append(item)
    return vendored_libs, paths


def remove_duplicate_vendored(vendor_dir, mf_vendored_names):
    intersec = set((f.name for f in vendor_dir.iterdir())).intersection(
        mf_vendored_names
    )
    delete_all(*[vendor_dir / p for p in intersec])
    return intersec


def fetch_licenses(*info_dir, vendor_dir):
    for file in chain.from_iterable(map(iter_subtree, info_dir)):
        if "LICENSE" in file.name:
            library = file.parent.name.split("-")[0]
            shutil.copy(file, vendor_dir / ("%s.LICENSE" % library))
        else:
            continue


def vendor(vendor_dir, mf_vendor_dir):
    # remove everything
    delete_all(*vendor_dir.iterdir(), whitelist=WHITELIST)

    exclude_subdirs = []
    # Iterate on the vendor*.txt files
    for vendor_file in glob.glob(f"{vendor_dir.name}/vendor*.txt"):
        # We extract the subdirectory we are going to extract into
        subdir = VENDOR_SUBDIR.match(vendor_file).group(1)
        # Includes "any" but it doesn't really matter unless you install "any"
        exclude_subdirs.append(subdir)

    for subdir in exclude_subdirs:
        create_init_file = False
        if subdir == "any":
            vendor_subdir = vendor_dir
            mf_vendor_subdir = mf_vendor_dir
            # target package is <parent>.<vendor_dir>; foo/vendor -> foo.vendor
            pkgname = f"{TL_PACKAGE_NAME}.{vendor_dir.parent.name}.{vendor_dir.name}"
        else:
            create_init_file = True
            vendor_subdir = vendor_dir / subdir
            mf_vendor_subdir = vendor_dir / subdir
            # target package is <parent>.<vendor_dir>; foo/vendor -> foo.vendor
            pkgname = f"{TL_PACKAGE_NAME}.{vendor_dir.parent.name}.{vendor_dir.name}.{vendor_subdir.name}"

        install_args = [
            "python3",
            "-m",
            "pip",
            "install",
            "-t",
            str(vendor_subdir),
            "--no-compile",
            "-r",
            "%s/vendor_%s.txt" % (str(vendor_dir), subdir),
        ]
        # We check in mf_vendor_dir
        if mf_vendor_subdir.is_dir():
            # We actually have stuff here so we figure out what we will need to exclude
            # from what we install
            excluded_installed = find_mf_vendored(
                *mf_vendor_subdir.iterdir(), exclude_dirs=frozenset(exclude_subdirs)
            )
            install_args.extend(
                ["-r", "%s/vendor_%s.txt" % (str(mf_vendor_dir), subdir)]
            )
        else:
            excluded_installed = []
        # install with pip
        subprocess.run(install_args)

        # fetch licenses
        fetch_licenses(*vendor_subdir.glob("*.dist-info"), vendor_dir=vendor_subdir)

        # delete stuff that's not needed
        delete_all(
            *vendor_subdir.glob("*.dist-info"),
            *vendor_subdir.glob("*.egg-info"),
            vendor_subdir / "bin",
        )

        # Touch a __init__.py file
        if create_init_file:
            with open(
                "%s/__init__.py" % str(vendor_subdir), "w+", encoding="utf-8"
            ) as f:
                f.write("# Empty file")

        mf_vendored_libs = remove_duplicate_vendored(vendor_subdir, excluded_installed)
        vendored_libs, paths = find_vendored_libs(
            vendor_subdir, WHITELIST, exclude_subdirs
        )

        replacements = []
        for lib in vendored_libs:
            replacements += (
                partial(  # import bar -> import foo.vendor.bar
                    re.compile(r"(^\s*)import {}\n".format(lib), flags=re.M).sub,
                    r"\1from {} import {}\n".format(pkgname, lib),
                ),
                partial(  # from bar -> from foo.vendor.bar
                    re.compile(r"(^\s*)from {}(\.|\s+)".format(lib), flags=re.M).sub,
                    r"\1from {}.{}\2".format(pkgname, lib),
                ),
            )
        for lib in mf_vendored_libs:
            replacements += (
                partial(  # import bar -> import foo.vendor.bar
                    re.compile(r"(^\s*)import {}\n".format(lib), flags=re.M).sub,
                    r"\1from {} import {}\n".format(
                        ".".join([MF_VENDOR_PREFIX, pkgname]), lib
                    ),
                ),
                partial(  # from bar -> from foo.vendor.bar
                    re.compile(r"(^\s*)from {}(\.|\s+)".format(lib), flags=re.M).sub,
                    r"\1from {}.{}\2".format(
                        ".".join([MF_VENDOR_PREFIX, pkgname]), lib
                    ),
                ),
            )

        for file in chain.from_iterable(map(iter_subtree, paths)):
            if file.suffix == ".py":
                patch_vendor_imports(file, replacements)


if __name__ == "__main__":
    here = Path("__file__").resolve().parent
    vendor_tl_dir = here / "vendor"
    has_vendor_file = len(glob.glob(f"{vendor_tl_dir.name}/vendor*.txt")) > 0
    assert has_vendor_file, "vendor/vendor*.txt file not found"

    # Locate the base vendor directory in the main metaflow distribution.
    assert len(sys.argv) == 2, (
        "Usage %s <path to vendor directory in metaflow>" % sys.argv[0]
    )
    mf_vendor = Path(sys.argv[1]).resolve()
    assert mf_vendor.is_dir(), "%s is not a directory" % str(mf_vendor)

    vendor(vendor_tl_dir, mf_vendor)



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import json
import os
import shutil
import tempfile

from itertools import chain, product
from typing import Dict, List, Optional, Tuple, cast

from urllib.parse import unquote, urlparse

from metaflow.debug import debug

# Consider re-adding if resolving fails weirdly.
# from metaflow.exception import retry_exp_backoff
from metaflow.metaflow_config import CONDA_SYS_MARKERS

from ..env_descr import (
    EnvType,
    PackageSpecification,  # noqa
    PypiCachePackage,
    PypiPackageSpecification,
    ResolvedEnvironment,
)

from ..pypi_package_builder import PackageToBuild, build_pypi_packages

from ..utils import (
    CondaException,
    arch_id,
    clean_up_double_equal,
    correct_splitext,
    filter_user_reqs_by_markers,
    get_glibc_version,
    parse_explicit_path_pypi,
    parse_explicit_url_pypi,
    pypi_tags_from_arch,
)
from . import Resolver

_FAKE_WHEEL = "_fake-1.0-py3-none-any.whl"


class PipResolver(Resolver):
    TYPES = ["pip"]
    REQUIRES_BUILDER_ENV = True

    # @retry_exp_backoff(
    #    (CondaException, OSError),
    #    "Failed to resolve pip environment",
    #    tries=3,
    #    max_delay_per_retry=5,
    # )
    def resolve(
        self,
        env_type: EnvType,
        python_version_requested: str,
        deps: Dict[str, List[str]],
        sources: Dict[str, List[str]],
        extras: Dict[str, List[str]],
        architecture: str,
        builder_envs: Optional[List[ResolvedEnvironment]] = None,
        base_env: Optional[ResolvedEnvironment] = None,
        file_paths: Dict[str, List[str]] = {},
        full_id_unique_keys: Dict[str, str] = {},
    ) -> Tuple[ResolvedEnvironment, Optional[List[ResolvedEnvironment]]]:
        if base_env:
            # For base environments, we may have built packages already so for those
            # packages, we need to make them available again to pip when resolving.
            # We therefore check packages that are:
            #   - not directly downloadable from their URL so, in this case, the URL is
            #     fake (ie: it contains FAKEURL_PATHCOMPONENT -- this is used when
            #     we build a package from a local directory/GIT repo, etc)
            #   - OR derived from a source package. This is the case when we build
            #     a wheel from a source package.
            #   - OR packages that have a download URL that is not a pypi URL or belongs to
            #     the sources we have for pypi. For instance pacakage @ github.com/..../package.wheel
            local_packages = [
                p
                for p in base_env.packages
                if p.TYPE == "pypi"
                and (
                    not p.is_downloadable_url()
                    or p.is_derived()
                    or p.is_external_url(sources)  # type: ignore[attr-defined]
                )
            ]
        else:
            local_packages = None
        # Some args may be two actual arguments like "-f <something>" thus the map
        extra_args = list(
            chain.from_iterable(
                map(lambda x: x.split(maxsplit=1), (e for e in extras.get("pypi", [])))
            )
        )

        if builder_envs is None:
            raise CondaException(
                "Cannot build a PYPI only environment without a builder "
                "environment. This is a bug -- please report"
            )
        # These can be the same if building for arch on same arch
        builder_env = [r for r in builder_envs if r.env_id.arch == arch_id()][0]
        base_conda_env = [r for r in builder_envs if r.env_id.arch == architecture][0]

        real_deps = filter_user_reqs_by_markers(
            {"pypi": deps.get("pypi", [])}, python_version_requested, architecture
        )["pypi"]
        # We get the python version for this builder env
        python_version = None  # type: Optional[str]
        for p in builder_env.packages:
            if p.filename.startswith("python-"):
                python_version = p.package_version
                break
        if python_version is None:
            raise CondaException(
                "Could not determine version of Python from conda packages"
            )

        # Create the environment in which we will call pip
        debug.conda_exec("Creating builder conda environment")
        builder_python = os.path.join(
            self._conda.create_builder_env(builder_env), "bin", "python"
        )
        # Add the path to the binary dir of the builder environment because
        # things in there may be needed to build/resolve the environment (one that has
        # come up is git-lfs)
        addl_env = {"PATH": os.path.dirname(builder_python) + ":" + os.environ["PATH"]}

        packages = []  # type: List[PackageSpecification]
        with tempfile.TemporaryDirectory() as pypi_dir:
            args = [
                "-m",
                "pip",
                "-v",
                "--isolated",
                "install",
                "--ignore-installed",
                "--dry-run",
                "--target",
                pypi_dir,
                "--report",
                os.path.join(pypi_dir, "out.json"),
            ]
            args.extend(extra_args)

            pypi_sources = sources.get("pypi", [])
            # The first source is always the index
            args.extend(["-i", pypi_sources[0]])
            for c in pypi_sources[1:]:
                args.extend(["--extra-index-url", c])

            if arch_id() == "linux-64":
                # Glibc version only relevant on linux
                platform_glibc = get_glibc_version()
                if platform_glibc is None:
                    raise CondaException("Could not determine the system GLIBC version")
            else:
                same_glibc = True

            if architecture == "linux-64":
                # Get the latest supported GLIBC version so we can generate
                # the proper tags
                glibc_versions_list = [
                    d for d in deps.get("sys", []) if d.startswith("__glibc")
                ]
                if len(glibc_versions_list) != 1:
                    raise CondaException("Could not determine maximum GLIBC version")
                # Get version looking like 2.27=0
                glibc_version = glibc_versions_list[0][len("__glibc==") :]
                # Strip =0
                glibc_version = glibc_version.split("=", 1)[0]

                if arch_id() == "linux-64":
                    same_glibc = glibc_version == platform_glibc

                # Replace . with _
                glibc_version = glibc_version.replace(".", "_")
            else:
                # It doesn't matter -- not used so don't compute
                glibc_version = ""
                same_glibc = True

            supported_tags = pypi_tags_from_arch(
                python_version, architecture, glibc_version
            )

            if architecture != arch_id() or not same_glibc:
                implementations = []  # type: List[str]
                abis = []  # type: List[str]
                platforms = []  # type: List[str]
                for tag in supported_tags:
                    implementations.append(tag.interpreter)
                    abis.append(tag.abi)
                    platforms.append(tag.platform)
                implementations = [x.interpreter for x in supported_tags]
                extra_args = [
                    "--only-binary=:all:",
                    # Seems to overly constrain stuff
                    # *(
                    #    chain.from_iterable(
                    #        product(["--implementation"], set(implementations))
                    #    )
                    # ),
                    *list(chain.from_iterable(product(["--abi"], set(abis)))),
                    *list(chain.from_iterable(product(["--platform"], set(platforms)))),
                ]

                args.extend(extra_args)

                # We will also override certain things when running pip -- particularly
                # platform.machine, etc which are used to compute the default_environment
                # that pip uses to evaluate markers. If we don't do this, resolving
                # an environment that needs torch from a mac for a linux box will fail
                # to include all dependencies because the markers specifying that
                # these additional packages are needed for linux are skipped as
                # pip uses the default_environment of the *current* machine and not the
                # target machine.
                addl_env["PIP_CUSTOMIZE_OVERRIDES"] = json.dumps(
                    CONDA_SYS_MARKERS.get(architecture, {})
                )
                addl_env["PYTHONPATH"] = os.path.join(
                    os.path.dirname(__file__), "..", "pip_customize"
                )
                debug.conda_exec(
                    "Launching pip with PYTHONPATH: %s and PIP_CUSTOMIZE_OVERRIDES: %s"
                    % (addl_env["PYTHONPATH"], addl_env["PIP_CUSTOMIZE_OVERRIDES"])
                )

            # If we have local packages, we download them to a directory and point
            # pip to it using the `--find-links` argument.
            local_packages_dict = {}  # type: Dict[str, PackageSpecification]
            if local_packages:
                # This is a bit convoluted but we try to avoid downloading packages
                # that we already have but we also don't want to point pip to a directory
                # full of packages that we may not want to use so we will create symlinks
                # to packages we already have and download the others

                # This is where we typically download/cache any pypi packages (source)
                base_local_pypi_packages = self._conda.package_dir("pypi")

                # This is where we will create symlinks in.
                tmp_local_pypi_packages = os.path.realpath(
                    os.path.join(pypi_dir, "local_packages", "pypi")
                )

                os.makedirs(tmp_local_pypi_packages)

                # Create this in case we need to move the local packages at some point
                os.makedirs(base_local_pypi_packages, exist_ok=True)

                new_local_packages = []  # type: List[PackageSpecification]
                for p in local_packages:
                    for fmt in PypiPackageSpecification.allowed_formats():
                        filename = "%s%s" % (p.filename, fmt)
                        if os.path.isfile(
                            os.path.join(base_local_pypi_packages, filename)
                        ):
                            os.symlink(
                                os.path.join(base_local_pypi_packages, filename),
                                os.path.join(tmp_local_pypi_packages, filename),
                            )
                            local_packages_dict[
                                os.path.join(tmp_local_pypi_packages, filename)
                            ] = p
                            p.add_local_file(
                                fmt,
                                os.path.join(base_local_pypi_packages, filename),
                                replace=True,
                            )
                            break
                    else:
                        new_local_packages.append(p)
                local_packages = new_local_packages
                # This will not fetch on the web so no need for auth object
                # Note we don't add pypi here because lazy_fetch_packages does that
                # automatically.
                self._conda.lazy_fetch_packages(
                    local_packages, None, os.path.join(pypi_dir, "local_packages")
                )
                args.extend(
                    ["--find-links", os.path.join(pypi_dir, "local_packages", "pypi")]
                )

                for p in local_packages:
                    for _, f in p.local_files:
                        local_packages_dict[os.path.realpath(f)] = p
                debug.conda_exec(
                    "Locally present files: %s" % ", ".join(local_packages_dict)
                )

            # Unfortunately, pip doesn't like things like ==<= so we need to strip
            # the ==
            args.extend(clean_up_double_equal(real_deps))

            self._conda.call_binary(args, binary=builder_python, addl_env=addl_env)

            # We should now have a json blob in out.json
            with open(
                os.path.join(pypi_dir, "out.json"), mode="r", encoding="utf-8"
            ) as json_file:
                desc = json.load(json_file)

            to_build_pkg_info = {}  # type: Dict[str, PackageToBuild]
            for package_desc in desc["install"]:
                # Three cases here:
                #  - we have a "dir_info" portion:
                #    - this means we have a directory
                #  - we have a "vcs_info" portion:
                #    - this means we are including a Github (for example) package directly
                #    - if we are the same arch as the target arch, we build the
                #      package and add it
                #  - we have an "archive_info" portion:
                #    - this means we have a direct URL to get the package from
                #    - if this is a wheel package, we add it directly
                #    - if this is not a wheel, we will build the package
                # The spec is given here:
                # https://packaging.python.org/en/latest/specifications/direct-url-data-structure/
                if debug.conda_exec:
                    package_desc = {
                        k: v
                        for k, v in package_desc.items()
                        if k in ("download_info", "vcs_info", "url", "subdirectory")
                    }
                    debug.conda_exec("Need to install %s" % str(package_desc))
                dl_info = package_desc["download_info"]
                url = dl_info["url"]
                # Extract hash if we have it
                # We have "archive_info"
                file_hash = None
                if "archive_info" in dl_info:
                    if "hashes" in dl_info["archive_info"]:
                        hashes = dl_info["archive_info"]["hashes"]
                        for fmt, val in hashes.items():
                            if fmt == PypiPackageSpecification.base_hash_name():
                                file_hash = "%s=%s" % (fmt, val)
                                break
                        else:
                            raise CondaException(
                                "Cannot find hash '%s' for package at '%s'"
                                % (
                                    PypiPackageSpecification.base_hash_name(),
                                    url,
                                )
                            )
                    else:
                        # Fallback on older "hash" field
                        file_hash = dl_info["archive_info"].get("hash")

                if "dir_info" in dl_info or url.startswith("file://"):
                    url = unquote(url)
                    to_build_local_pkg = None  # type: Optional[str]
                    local_path = url[7:]
                    if "dir_info" in dl_info and dl_info["dir_info"].get(
                        "editable", False
                    ):
                        raise CondaException(
                            "Cannot include an editable PYPI package: '%s'" % url
                        )
                    if os.path.isdir(local_path):
                        if os.path.isfile(os.path.join(local_path, "setup.py")):
                            package_name, package_version = (
                                self._conda.call_binary(
                                    [
                                        os.path.join(local_path, "setup.py"),
                                        "-q",
                                        "--name",
                                        "--version",
                                    ],
                                    binary=builder_python,
                                    addl_env=addl_env,
                                )
                                .decode(encoding="utf-8")
                                .splitlines()
                            )
                        elif os.path.isfile(os.path.join(local_path, "pyproject.toml")):
                            package_name, package_version = (
                                self._conda.call_binary(
                                    [
                                        "-c",
                                        "import tomli; f = open('%s', mode='rb'); "
                                        "d = tomli.load(f); "
                                        "print(d.get('project', d.get('poetry', d['tool']['poetry']))['name']); "
                                        "print(d.get('project', d.get('poetry', d['tool']['poetry']))['version'])"
                                        % os.path.join(local_path, "pyproject.toml"),
                                    ],
                                    binary=builder_python,
                                    addl_env=addl_env,
                                )
                                .decode(encoding="utf-8")
                                .splitlines()
                            )
                        else:
                            raise CondaException(
                                "Local directory '%s' is not supported as it is "
                                "missing a 'setup.py' or 'pyproject.toml'" % local_path
                            )
                        # Make sure to use the quoted URL here
                        to_build_local_pkg = os.path.join(
                            dl_info["url"],
                            "%s-%s.whl" % (package_name, package_version),
                        )
                    else:
                        # A local wheel or tarball
                        if local_path in local_packages_dict:
                            # We are going to move this file to a less "temporary"
                            # location so that it can be installed if needed

                            pkg_spec = local_packages_dict[local_path]
                            if not os.path.islink(local_path):
                                # If this is not a link, it means we downloaded it
                                debug.conda_exec("Known package -- moving in place")
                                filename = os.path.split(local_path)[1]
                                file_format = correct_splitext(filename)[1]
                                shutil.move(local_path, self._conda.package_dir("pypi"))

                                # Update where the local file resides for this package
                                pkg_spec.add_local_file(
                                    file_format,
                                    os.path.join(
                                        self._conda.package_dir("pypi"), filename
                                    ),
                                    replace=True,
                                )
                            else:
                                debug.conda_exec("Known package already in place")
                            packages.append(pkg_spec)
                        else:
                            parse_result = parse_explicit_path_pypi(
                                url,
                                hash_value=(
                                    file_hash.split("=")[1] if file_hash else None
                                ),
                            )
                            if parse_result.url_format != ".whl":
                                # This is a source package so we need to build it
                                to_build_local_pkg = dl_info["url"]
                            else:
                                package_spec = PypiPackageSpecification(
                                    parse_result.filename,
                                    parse_result.url,
                                    is_real_url=False,
                                    url_format=parse_result.url_format,
                                )
                                # We extract the actual local file so we can use that for now
                                # Note that the url in PypiPackageSpecication is a fake
                                # one that looks like file://local-<hash>/... which is meant
                                # to act as a key for the package.
                                package_spec.add_local_file(
                                    parse_result.url_format, url[7:]
                                )
                                packages.append(package_spec)
                    if to_build_local_pkg:
                        # Note the hash value here is the hash of the *source* tarball
                        # This is fine as it still uniquely identifies the package
                        # we will be building.
                        parse_result = parse_explicit_path_pypi(
                            to_build_local_pkg,
                            hash_value=file_hash.split("=")[1] if file_hash else None,
                        )
                        cache_base_url = PypiCachePackage.make_partial_cache_url(
                            parse_result.url, is_real_url=False
                        )

                        to_build_pkg_info[cache_base_url] = PackageToBuild(
                            dl_info["url"],
                            # We get the name once we build the package
                            PypiPackageSpecification(
                                _FAKE_WHEEL,
                                parse_result.url,
                                is_real_url=False,
                                url_format=".whl",
                            ),
                        )
                elif "vcs_info" in dl_info:
                    # This is a repository that we can go build

                    # Create the URL
                    base_build_url = "%s+%s@%s" % (
                        dl_info["vcs_info"]["vcs"],
                        dl_info["url"],
                        dl_info["vcs_info"]["commit_id"],
                    )
                    # We form a "fake" URL which will give us a unique key so we can
                    # look up the package build in the cache. Given we have the
                    # commit_id, we assume that for a combination of repo, commit_id
                    # and subdirectory, we have a uniquely built package. This will
                    # give users some consistency as well in the sense that an
                    # environment that uses the same git package reference will
                    # actually use the same package
                    base_pkg_url = "%s/%s" % (
                        dl_info["url"],
                        dl_info["vcs_info"]["commit_id"],
                    )
                    if "subdirectory" in dl_info:
                        base_build_url += "#subdirectory=%s" % dl_info["subdirectory"]
                        base_pkg_url += "/%s" % dl_info["subdirectory"]
                    cache_base_url = PypiCachePackage.make_partial_cache_url(
                        base_pkg_url, is_real_url=False
                    )
                    to_build_pkg_info[cache_base_url] = PackageToBuild(
                        base_build_url,
                        # We get the name once we build the package
                        PypiPackageSpecification(
                            _FAKE_WHEEL,
                            base_pkg_url,
                            is_real_url=False,
                            url_format=".whl",
                        ),
                    )
                else:
                    # Here we have archive_info
                    parse_result = parse_explicit_url_pypi(
                        "%s#%s" % (url, file_hash) if file_hash is not None else url
                    )
                    if parse_result.url_format != ".whl":
                        # We need to build non wheel files.
                        url_parse_result = urlparse(cast(str, dl_info["url"]))

                        is_real_url = False
                        if url_parse_result.scheme == "file":
                            raise CondaException(
                                "URL %s should have been identified as a local file"
                                % dl_info["url"]
                            )
                        else:
                            cache_base_url = PypiCachePackage.make_partial_cache_url(
                                parse_result.url, is_real_url=True
                            )
                            is_real_url = True

                        spec = PypiPackageSpecification(
                            parse_result.filename,
                            parse_result.url,
                            is_real_url=is_real_url,
                            url_format=parse_result.url_format,
                        )
                        to_build_pkg_info[cache_base_url] = PackageToBuild(
                            dl_info["url"],
                            spec,
                        )
                        if not is_real_url:
                            # We have a local file for this tar-ball so we need to
                            # indicate where it is.
                            spec.add_local_file(
                                parse_result.url_format, url_parse_result.path
                            )
                            to_build_pkg_info[cache_base_url].have_formats.append(
                                parse_result.url_format
                            )
                    else:
                        # This is a wheel so we can just add it.
                        packages.append(
                            PypiPackageSpecification(
                                parse_result.filename,
                                parse_result.url,
                                url_format=parse_result.url_format,
                                hashes=(
                                    {parse_result.url_format: parse_result.hash}
                                    if parse_result.hash
                                    else None
                                ),
                            )
                        )
            if to_build_pkg_info:
                if not self._conda.storage:
                    raise CondaException(
                        "Cannot create a relocatable environment as it depends on "
                        "local files or non wheels and no storage backend is defined: %s"
                        % ", ".join([v.url for v in to_build_pkg_info.values()])
                    )

                built_pypi_packages, builder_envs = build_pypi_packages(
                    self._conda,
                    self._conda.storage,
                    python_version,
                    to_build_pkg_info,
                    builder_envs,
                    pypi_dir,
                    architecture,
                    supported_tags,
                    sources.get("pypi", []),
                )

                packages.extend(built_pypi_packages)

        # packages has pypi only packages so order doesn't really matter (we can just
        # directly append))
        packages += list(base_conda_env.packages)
        return (
            ResolvedEnvironment(
                deps,
                sources,
                extras,
                architecture,
                all_packages=packages,
                env_type=env_type,
                # full_id_unique_keys is a uv specific cache invalidation mechanism,
                # and PipResolver shouldn't need it. We still pass this along as {}
                # to make code consistent.
                full_id_unique_keys=full_id_unique_keys,
            ),
            builder_envs,
        )



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
from __future__ import annotations

import os
import shutil

from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)  # noqa

if TYPE_CHECKING:
    import metaflow.datastore.datastore_storage
    import metaflow_extensions.nflx.plugins.conda.conda

from metaflow.debug import debug

from metaflow._vendor.packaging.tags import Tag
from metaflow._vendor.packaging.utils import parse_wheel_filename

from .env_descr import (
    EnvType,
    PackageSpecification,
    PypiCachePackage,
    PypiPackageSpecification,
    ResolvedEnvironment,
)

from .resolvers.builder_envs_resolver import BuilderEnvsResolver

from .utils import (
    CondaException,
    arch_id,
    auth_from_urls,
    change_pypi_package_version,
    correct_splitext,
    parse_explicit_path_pypi,
    version_to_str,
)

_DEV_TRANS = str.maketrans("abcdef", "012345")


# This is a dataclass -- can move to that when we only support 3.7+
class PackageToBuild:
    def __init__(
        self,
        url: str,
        spec: Optional[PackageSpecification] = None,
        have_formats: Optional[List[str]] = None,
    ):
        self.url = url
        self.spec = spec
        self.have_formats = have_formats or []


def build_pypi_packages(
    conda: "metaflow_extensions.nflx.plugins.conda.conda.Conda",
    storage: "metaflow.datastore.datastore_storage.DataStoreStorage",
    python_version: str,
    to_build_pkg_info: Dict[str, PackageToBuild],
    builder_envs: Optional[List[ResolvedEnvironment]],
    build_dir: str,
    architecture: str,
    supported_tags: List[Tag],
    pypi_sources: List[str],
) -> Tuple[List[PackageSpecification], Optional[List[ResolvedEnvironment]]]:
    # We check in the cache -- we don't actually have the filename or
    # hash so we check things starting with the partial URL.
    # The URL in cache will be:
    #  - <base url>/<filename>/<hash>/<filename>

    debug.conda_exec(
        "Checking for pre-built packages: %s"
        % ", ".join(["%s @ %s" % (v.spec, k) for k, v in to_build_pkg_info.items()])
    )
    found_files = cast(
        Sequence[Tuple[str, bool]], storage.list_content(to_build_pkg_info.keys())
    )

    keys_to_check = set()  # type: Set[str]

    # Key: key in to_build_pkg_info
    # Value: list of possible cache paths
    possible_wheels = {}  # type: Dict[str, List[str]]
    for cache_path, is_file in found_files:
        cache_path = cache_path.rstrip("/")
        if is_file:
            raise CondaException("Invalid cache content at '%s'" % cache_path)
        base_cache_path, cache_filename_with_ext = os.path.split(cache_path)
        cache_format = correct_splitext(cache_filename_with_ext)[1]
        if cache_format != ".whl":
            # This is a source format -- we add it to the keys_to_check so we can
            keys_to_check.add(cache_path)
            debug.conda_exec("Found source package at '%s'" % cache_path)
        else:
            # There may be multiple wheel files so we want to pick the best one
            # so we record for now and then we will pick the best one.
            possible_wheels.setdefault(base_cache_path, []).append(cache_path)
            debug.conda_exec("Found potential pre-built package at '%s'" % cache_path)

    # We now check and pick the best wheel if one is compatible and then we will
    # check it further
    for key, wheel_potentials in possible_wheels.items():
        for t in supported_tags:
            # Tags are ordered from most-preferred to least preferred
            for p in wheel_potentials:
                # Potentials are in no particular order but we will
                # effectively get a package with the most preferred tag
                # if one exists
                wheel_name = os.path.split(p)[1]
                _, _, _, tags = parse_wheel_filename(wheel_name)
                if t in tags:
                    keys_to_check.add(p)
                    debug.conda_exec("%s: matching package @ %s" % (key, p))
                    break
            else:
                # If we don't find a match, continue to next tag (and
                # skip break of outer loop on next line)
                continue
            break

    # We now check for hashes for those packages we did find (it's the
    # next level down in the cache)
    found_files = cast(Sequence[Tuple[str, bool]], storage.list_content(keys_to_check))
    for cache_path, is_file in found_files:
        cache_path = cache_path.rstrip("/")
        if is_file:
            raise CondaException("Invalid cache content at '%s'" % cache_path)
        head, _ = os.path.split(cache_path)
        base_cache_path, cache_filename_with_ext = os.path.split(head)
        cache_filename, cache_format = correct_splitext(cache_filename_with_ext)

        pkg_info = to_build_pkg_info[base_cache_path]
        pkg_spec = cast(PypiPackageSpecification, pkg_info.spec)
        pkg_info.have_formats.append(cache_format)
        if cache_format == ".whl":
            # In some cases, we don't know the filename so we change it here (or
            # we need to update it since a tarball has a generic name without
            # ABI, etc but a wheel name has more information)
            if pkg_spec.filename != cache_filename:
                pkg_spec = cast(
                    PypiPackageSpecification,
                    pkg_spec.clone_with_filename(cache_filename),
                )
                pkg_info.spec = pkg_spec
        debug.conda_exec(
            "%s:%s adding cache file %s"
            % (
                pkg_spec.filename,
                cache_format,
                os.path.join(cache_path, cache_filename_with_ext),
            )
        )
        pkg_spec.add_cached_version(
            cache_format,
            PypiCachePackage(os.path.join(cache_path, cache_filename_with_ext)),
        )

    # Determine what we need to build -- all non wheels
    # We try to build even across architecture but will check if we can use it across
    # architectures after (ie: if the package is noarch).
    keys_to_build = [
        k for k, v in to_build_pkg_info.items() if ".whl" not in v.have_formats
    ]

    if not keys_to_build:
        return [v.spec for v in to_build_pkg_info.values() if v.spec], builder_envs

    debug.conda_exec(
        "Going to build packages %s"
        % ", ".join(
            [
                to_build_pkg_info[k].spec.filename  # type: ignore[union-attr]
                for k in keys_to_build
                if to_build_pkg_info[k].spec is not None
            ]
        )
    )
    # Here we are the same architecture so we can go ahead and build the wheel and
    # add it.
    conda.echo(" (building PYPI packages from repositories)", nl=False)
    debug.conda_exec("Creating builder environment to build PYPI packages")

    # Create the environment in which we will call pip
    # We look for a builder environment for the architecture we are building on
    # (we never build cross arch)
    builder_env = None  # type: Optional[ResolvedEnvironment]
    if builder_envs is not None:
        t = [r for r in builder_envs if r.env_id.arch == arch_id()]
        if t:
            builder_env = t[0]

    if not builder_env:
        builder_env, builder_envs = BuilderEnvsResolver(conda).resolve(
            EnvType.CONDA_ONLY,
            f"python=={python_version}",
            {"conda": ["python==%s" % python_version]},
            {},
            {},
            arch_id(),
            builder_envs,
            None,
        )

        if builder_envs:
            builder_envs.append(builder_env)
        else:
            builder_envs = [builder_env]

    builder_python = os.path.join(
        conda.create_builder_env(builder_env), "bin", "python"
    )

    # Download any source either from cache or the web. We can use our typical
    # lazy fetch to do this. We just make sure that we only pass it packages that
    # it has something to fetch
    target_directory = conda.package_dir("pypi")
    os.makedirs(target_directory, exist_ok=True)
    pkgs_to_fetch = cast(
        List[PypiPackageSpecification],
        [to_build_pkg_info[k].spec for k in keys_to_build],
    )
    pkgs_to_fetch = list(
        filter(
            lambda x: x.is_downloadable_url() or x.cached_version(x.url_format),
            pkgs_to_fetch,
        )
    )
    debug.conda_exec(
        "Going to fetch sources for %s" % ", ".join([p.filename for p in pkgs_to_fetch])
    )
    if pkgs_to_fetch:
        conda.lazy_fetch_packages(
            pkgs_to_fetch, auth_from_urls(pypi_sources), target_directory
        )

    # We try to build and will check if the tags are acceptable after the fact. This
    # may be a bit slower than the reverse but it allows us to have more packages built
    # (for example noarch packages) across architectures so it feels like a fair tradeoff.
    # Ideally we would be able to determine this a-priori but I haven't found a way to
    # do that.
    with ThreadPoolExecutor() as executor:
        build_result = [
            executor.submit(
                _build_with_pip,
                conda,
                builder_python,
                os.path.join(build_dir, "build_%d" % idx),
                key,
                cast(PypiPackageSpecification, to_build_pkg_info[key].spec),
                to_build_pkg_info[key].url,
            )
            for idx, key in enumerate(keys_to_build)
        ]

        unsupported_wheel = []  # type: List[str]
        for f in as_completed(build_result):
            key, build_dir = f.result()
            wheel_files = [
                f
                for f in os.listdir(build_dir)
                if os.path.isfile(os.path.join(build_dir, f)) and f.endswith(".whl")
            ]
            if len(wheel_files) != 1:
                raise CondaException(
                    "Could not build '%s' -- found built packages: %s"
                    % (key, wheel_files)
                )

            wheel_file = os.path.join(build_dir, wheel_files[0])

            # Technically we only need to check the tags for packages that are cross-arch
            # built but determining what is cross-arch would also require looking at
            # glibc version so this seems cheap enough to not complicate things.
            wheel_tags = parse_wheel_filename(os.path.basename(wheel_file))[3]
            if len(wheel_tags.intersection(supported_tags)) == 0:
                debug.conda_exec(
                    "Package '%s' was built with %s but target architecture "
                    "supports %s" % (key, wheel_tags, supported_tags)
                )
                unsupported_wheel.append(key)
                continue

            pkg_spec = cast(PypiPackageSpecification, to_build_pkg_info[key].spec)
            # Move the built wheel to a less temporary location
            wheel_file = shutil.copy(wheel_file, target_directory)

            parse_result = parse_explicit_path_pypi("file://%s" % wheel_file)
            # If the source is not an actual URL, we are going to change the name
            # of the package to avoid any potential conflict. We consider that
            # packages derived from internet URLs (so likely a source package)
            # do not need name changes
            if not pkg_spec.is_downloadable_url():
                pkg_version = parse_wheel_filename(parse_result.filename + ".whl")[1]

                pkg_dev = pkg_version.dev
                pkg_local = pkg_version.local
                if not pkg_dev:
                    wheel_hash = PypiPackageSpecification.hash_pkg(wheel_file)

                    pkg_dev = int(wheel_hash[:8].translate(_DEV_TRANS))
                if pkg_local:
                    pkg_local += "_mfbuild"
                else:
                    pkg_local = "mfbuild"
                pkg_version_str = version_to_str(
                    pkg_version, {"dev": pkg_dev, "local": pkg_local}
                )
                wheel_file = change_pypi_package_version(
                    builder_python, wheel_file, pkg_version_str
                )
                parse_result = parse_explicit_path_pypi("file://%s" % wheel_file)

            debug.conda_exec("Package for '%s' built in '%s'" % (key, wheel_file))

            # We update because we need to change the filename mostly so that it
            # now reflects the abi, etc and all that goes in a wheel filename.
            pkg_spec = cast(
                PypiPackageSpecification,
                pkg_spec.clone_with_filename(parse_result.filename),
            )
            to_build_pkg_info[key].spec = pkg_spec
            pkg_spec.add_local_file(".whl", wheel_file, pkg_hash=parse_result.hash)

        if unsupported_wheel:
            raise CondaException(
                "Some built pypi packages are not compatible with the target "
                "architecture. This can happen when some dependencies do not have "
                "an available wheel for the target architecture and you are resolving "
                "an environment for a different architecture than the one you are on. "
                "Requirements would have us build: %s" % ", ".join(unsupported_wheel)
            )

    return [v.spec for v in to_build_pkg_info.values() if v.spec], builder_envs


def _build_with_pip(
    conda: "metaflow_extensions.nflx.plugins.conda.conda.Conda",
    binary: str,
    dest_path: str,
    key: str,
    spec: PypiPackageSpecification,
    build_url: str,
):
    src = spec.local_file(spec.url_format) or build_url
    debug.conda_exec("%s: building from '%s' in '%s'" % (key, src, dest_path))

    addl_env = {"PATH": os.path.dirname(binary) + ":" + os.environ["PATH"]}
    conda.call_binary(
        [
            "-m",
            "pip",
            "--isolated",
            "wheel",
            "--no-deps",
            "--progress-bar",
            "off",
            "-w",
            dest_path,
            src,
        ],
        binary=binary,
        addl_env=addl_env,
    )
    return key, dest_path



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import json
import os
import sys

from typing import Any, Dict, List, Set, TYPE_CHECKING, Tuple, Union, Optional, cast

from metaflow._vendor import click

from metaflow.exception import CommandException
from metaflow.metaflow_config import CONDA_PREFERRED_FORMAT, CONDA_TEST

from .conda.conda import Conda
from .conda.conda_common_decorator import StepRequirement
from .conda.conda_environment import CondaEnvironment
from .conda.envsresolver import EnvsResolver
from .conda.env_descr import (
    AliasType,
    EnvID,
    ResolvedEnvironment,
    MetaflowResolvedEnvironment,
)
from .conda.utils import dict_to_tstr, plural_marker, resolve_env_alias

if TYPE_CHECKING:
    import metaflow

# Very simple functions to create some highlights without importing other packages
_BOLD = 1
_ITALIC = 3
_UNDERLINE = 4
_RED = 31
_GREEN = 32
_YELLOW = 33
_BLUE = 34

# reqid->fullid->[paths]
_all_local_instances = None  # type: Optional[Dict[str, Dict[str, List[str]]]]


# This is a dataclass -- can move to that when we only support 3.7+
class StepEnvInfo:
    """
    Information about the environment needed for a step.

    Attributes
    ----------
    is_disabled : bool
        If True, environment management is disabled for the step

    error : Optional[str]
        If the specification of the environment is invalid, this will contain the
        reason for it. Note that the absence of an error does not guarantee that the
        environment will be fully resolved correctly (this is un-knowable in all cases).

    requested_hash : Optional[str]
        Hash of the requested environment. This is used to uniquely identify the requirements
        set by the user. Note that for a given requested hash, multiple environments may
        be created. A unique environment is determined by both its requested hash as well
        as its full hash. The full hash is only available for resolved environments.
        Will be none if is_disabled or error is set. If the environment is resolved then
        env.env_id.req_id will be the same as requested_hash.

    requested_packages : Dict[str, Dict[str, str]]
        Packages requested for the step. The first key is the type of package (pypi,
        conda or sys), the second key is the package name and the value is the version
        constraint.

    requested_sources : Dict[str, List[str]]
        Sources requested for the step. The first key is the type of source (conda or pypi)
        and the value is a list of the sources. The order of the list is preserved
        based on what the user indicated.

    from_env_name : Optional[str]
        If non-None, the environment is derived from another environment (either a named
        environment or a step's environment).

    is_fetch_at_exec : bool
        If True, the environment does not need to be known when deploying the flow
        but only at runtime.

    env : Optional[MetaflowResolvedEnvironment]
        If the environment is known an resolved, it will be returned here. Otherwise,
        this value will be None. This means that the environment will need to be
        resolved prior to deploying unless is_fetch_at_exec is True.
    """

    def __init__(
        self,
        is_disabled: bool,
        error: Optional[str] = None,
        requested_hash: Optional[str] = None,
        requested_packages: Optional[Dict[str, Dict[str, str]]] = None,
        requested_sources: Optional[Dict[str, List[str]]] = None,
        from_env_name: Optional[str] = None,
        is_fetch_at_exec: bool = False,
        env: Optional[MetaflowResolvedEnvironment] = None,
    ):
        self.is_disabled = is_disabled
        self.error = error
        self.requested_hash = requested_hash
        self.requested_packages = requested_packages or {}
        self.requested_sources = requested_sources or {}
        self.from_env_name = from_env_name
        self.is_fetch_at_exec = is_fetch_at_exec
        self.env = env

    def to_dict(self) -> Dict[str, Any]:
        return {
            "is_disabled": self.is_disabled,
            "error": self.error,
            "requested_hash": self.requested_hash,
            "requested_packages": self.requested_packages,
            "requested_sources": self.requested_sources,
            "from_env_name": self.from_env_name,
            "is_fetch_at_exec": self.is_fetch_at_exec,
            "env_id": self.env.env_id if self.env else None,
            "env": self.env.to_dict() if self.env else None,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "StepEnvInfo":
        return cls(
            data["is_disabled"],
            data["error"],
            data["requested_hash"],
            data["requested_packages"],
            data["requested_sources"],
            data["from_env_name"],
            data["is_fetch_at_exec"],
            (
                MetaflowResolvedEnvironment(
                    ResolvedEnvironment.from_dict(EnvID(*data["env_id"]), data["env"])
                )
                if data["env"]
                else None
            ),
        )


class RunnerCLI(object):
    name = "environment"

    def __init__(self, runner: "metaflow.Runner"):
        """
        Environment command for the flow. Requires the conda environment.
        """
        self.runner = runner

    def resolve(
        self,
        *args: str,
        arch: Optional[Union[str, List[str]]] = None,
        alias: Optional[List[str]] = None,
        force: bool = False,
        dry_run: bool = False,
        timeout: int = 1200,
    ) -> Dict[str, Dict[str, Tuple[MetaflowResolvedEnvironment, bool]]]:
        """
        Resolve the environments needed for this flow. Returns a dictionary associating
        the name of the step with the resolved environment.

        Note that steps that have conda disabled will not be present in the returned
        dictionary.

        Parameters
        ----------
        *args : str
            Name of the steps to resolve. If not specified, all steps are resolved
        arch : Union[str, List[str]], optional, default None
            Architecture(s) to resolve the environments for. If None, the default
            architecture is used (dependent on where the step should run)
        alias : List[str], optional, default None
            If specified, aliases to alias the resolved environment by. Only valid if
            only one environment is being resolved (ie: only one step)
        force : bool, default False
            Force the resolution of all environments even if they are already resolved.
        dry_run : bool, default False
            Perform the resolution but do not cache or write out the resolved environments.
        timeout : int, default 1200
            Timeout in seconds -- note that this effectively caps the total amount of time
            available for environment resolution.

        Returns
        -------
        Dict[str, Dict[str, Tuple[MetaflowResolvedEnvironment, bool]]]
            Resolved environments for all steps requested.
            The outer key is the name of the step
            The inner key is the architecture
            The element is a tuple consisting of the resolved environment and a boolean
            indicating if the environment was newly resolved or not (True if just
            resolved or False if previously resolved)
        """
        from metaflow.runner.utils import (
            handle_timeout,
            make_process_error_message,
            temporary_fifo,
        )

        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            command = (
                self.runner.api(**self.runner.top_level_kwargs)
                .environment()
                .resolve(
                    arch=arch,
                    alias=alias,
                    force=force,
                    dry_run=dry_run,
                    output=attribute_file_path,
                    steps_to_resolve=args if args else None,
                )
            )
            pid = self.runner.spm.run_command(
                [sys.executable, *command],
                env=self.runner.env_vars,
                cwd=self.runner.cwd,
                show_output=False,
            )
            command_obj = self.runner.spm.get(pid)
            if not command_obj:
                raise RuntimeError("Failed to get command object")

            content = handle_timeout(attribute_file_fd, command_obj, timeout)
            command_obj.sync_wait()

            if command_obj.process.returncode == 0:
                # We parse content and return the resolved environments
                content = json.loads(content)
                return {
                    k: {
                        kk: (
                            MetaflowResolvedEnvironment(
                                ResolvedEnvironment.from_dict(
                                    EnvID(vv["req_id"], vv["full_id"], kk),
                                    vv["env"],
                                )
                            ),
                            vv["newly_resolved"],
                        )
                        for kk, vv in v.items()
                    }
                    for k, v in content.items()
                }
            raise RuntimeError(make_process_error_message(command_obj))

    def show(
        self, *args: str, local_only: bool = False, timeout: int = 60
    ) -> Dict[str, StepEnvInfo]:
        """
        Returns information about the environments needed for the steps in the flow.

        For each requested step, information about the packages and sources requested
        as well as an optional resolved environment is returned.

        Parameters
        ----------
        args : str
            Steps to show information for. If none are specified, information for all
            steps will be returned.
        local_only : bool, default False
            If True, will only list environments as known if they are locally known. This
            primarily affects named environments or if you have
            METAFLOW_CONDA_USE_REMOTE_LATEST set to something different than ":none:".
        timeout : int, default 60
            Timeout in seconds for this call.

        Returns
        -------
        Dict[str, StepEnvInfo]
            For each step requested, returns information about the environment needed for
            that step.
        """
        from metaflow.runner.utils import (
            handle_timeout,
            make_process_error_message,
            temporary_fifo,
        )

        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            command = (
                self.runner.api(**self.runner.top_level_kwargs)
                .environment()
                .show(
                    local_only=local_only,
                    output=attribute_file_path,
                    steps_to_show=args if args else None,
                )
            )
            pid = self.runner.spm.run_command(
                [sys.executable, *command],
                env=self.runner.env_vars,
                cwd=self.runner.cwd,
                show_output=False,
            )
            command_obj = self.runner.spm.get(pid)
            assert command_obj
            content = handle_timeout(attribute_file_fd, command_obj, timeout)
            command_obj.sync_wait()

            if command_obj.process.returncode == 0:
                # We parse content and return the resolved environments
                content = json.loads(content)
                return {k: StepEnvInfo.from_dict(v) for k, v in content.items()}

            raise RuntimeError(make_process_error_message(command_obj))


@click.group()
def cli():
    pass


@cli.group(help="Commands related to environment manipulation.")
@click.pass_obj
def environment(obj):
    if obj.environment.TYPE != "conda":
        raise CommandException("'environment' requires a Conda environment")


@environment.command(help="Resolve environments for steps in a flow")
@click.option(
    "--alias",
    default=None,
    multiple=True,
    help="Alias the resolved environment; can be name or name:tag",
)
@click.option(
    "--force/--no-force",
    show_default=True,
    default=False,
    help="Force re-resolution of already resolved environments",
)
@click.option(
    "--dry-run/--no-dry-run",
    show_default=True,
    default=False,
    help="Dry-run -- only resolve, do not download, cache, persist or alias anything",
)
@click.option(
    "--arch",
    default=None,
    multiple=True,
    help="Architecture to resolve for. Can be specified multiple times. "
    "If not specified, defaults to the architecture for the step",
    required=False,
)
@click.option(
    "--output",
    show_default=True,
    default=None,
    type=click.Path(file_okay=True, writable=True, resolve_path=True),
    help="If specified, the environments for the requested steps are written out "
    "to this file. Note that this includes only newly resolved environments.",
)
@click.argument(
    "steps-to-resolve",
    required=False,
    nargs=-1,
)
@click.pass_obj
def resolve(
    obj: Any,
    steps_to_resolve: Tuple[str],
    arch: Optional[Tuple[str]] = None,
    alias: Optional[Tuple[str]] = None,
    force: bool = False,
    dry_run: bool = False,
    output: Optional[str] = None,
):

    conda = Conda(obj.echo, obj.flow_datastore.TYPE)
    resolver = EnvsResolver(conda)

    def _write_out_environments():
        # If we have an output file, write out to it
        if output:
            resolved_per_step = {}  # type: Dict[str, Dict[str, ResolvedEnvironment]]
            new_env_ids = set(env.env_id for _, env, _ in resolver.new_environments())
            for env_id, env, steps in resolver.resolved_environments():
                for step in steps:
                    resolved_per_step.setdefault(step, {})[env_id.arch] = env
            out_json = {}
            with open(output, "w", encoding="utf-8") as f:
                for step, envs in resolved_per_step.items():
                    for env_arch, env in envs.items():
                        out_json.setdefault(step, {})[env_arch] = {
                            "req_id": env.env_id.req_id,
                            "full_id": env.env_id.full_id,
                            "newly_resolved": env.env_id in new_env_ids,
                            "env": env.to_dict(),
                        }
                json.dump(out_json, f, indent=2)

    for step in obj.flow:
        if not steps_to_resolve or step.name in steps_to_resolve:
            for resolve_arch in arch or [None]:
                (
                    _,
                    step_arch,
                    req,
                    base_env,
                ) = CondaEnvironment.extract_merged_reqs_for_step(
                    conda,
                    obj.flow,
                    obj.flow_datastore.TYPE,
                    step,
                    override_arch=resolve_arch,
                )
                if req.is_disabled or req.is_fetch_at_exec:
                    break  # No point resolving anything else for this step
                base_from_full_id = False
                if req.from_env_name:
                    base_from_full_id = (
                        resolve_env_alias(req.from_env_name)[0] == AliasType.FULL_ID
                    )
                resolver.add_environment(
                    step_arch,
                    req.packages_as_str,
                    req.sources,
                    {},
                    step.name,
                    base_env,
                    base_from_full_id=base_from_full_id,
                    force=force,
                )

    per_req_id = {}  # type: Dict[str, Set[str]]

    if alias:
        aliases = list(alias)
    else:
        aliases = None
    for env_id, steps in resolver.non_resolved_environments():
        per_req_id.setdefault(env_id.req_id, set()).update(steps)
    if len(per_req_id) == 0:
        # Nothing to do but we may still need to alias
        if aliases is not None and not dry_run:
            all_resolved_envs = list(resolver.resolved_environments())
            if len(all_resolved_envs) == 0:
                obj.echo("No environments to resolve")
            elif len(all_resolved_envs) == 1:
                obj.echo("Environment already resolved -- aliasing only")
                conda.alias_environment(all_resolved_envs[0][0], aliases)
                conda.write_out_environments()
            else:
                raise CommandException(
                    "Cannot specify aliases if more than one environments to alias "
                    "-- found %d environments: one for each of steps %s"
                    % (
                        len(all_resolved_envs),
                        ", and ".join(", ".join(v[2]) for v in all_resolved_envs),
                    )
                )
        else:
            obj.echo("No environments to resolve, use --force to force re-resolution")
        _write_out_environments()
        return
    if aliases is not None and len(per_req_id) > 1:
        raise CommandException(
            "Cannot specify aliases if more than one environment to resolve "
            "-- found %d environments: one for each of steps %s"
            % (
                len(per_req_id),
                ", and ".join(", ".join(s) for s in per_req_id.values()),
            )
        )

    resolver.resolve_environments(obj.echo)
    # We group by req_id which will effectively be by set of steps. We only
    # print newly resolved environments
    resolved_per_req_id = {}  # type: Dict[str, Tuple[ResolvedEnvironment, Set[str]]]
    for env_id, env, steps in resolver.resolved_environments():
        v = resolved_per_req_id.setdefault(env_id.req_id, (env, set()))
        v[1].update(steps)
        if aliases and not dry_run:
            # We know this executes only once since we already checked that if tag
            # is set, there is only one environment being resolved
            conda.alias_environment(env.env_id, aliases)

    _compute_local_instances(conda)
    assert _all_local_instances is not None

    _write_out_environments()

    if obj.is_quiet:
        for req_id, (env, steps) in resolved_per_req_id.items():
            obj.echo_always(",".join(steps))
            obj.echo_always(
                env.quiet_print(
                    _all_local_instances.get(req_id, {}).get(env.env_id.full_id)
                )
            )
    else:
        for req_id, (env, steps) in resolved_per_req_id.items():
            obj.echo(
                "### Environment for step%s *%s* ###"
                % (plural_marker(len(steps)), ", ".join(steps))
            )
            obj.echo(
                env.pretty_print(
                    _all_local_instances.get(req_id, {}).get(env.env_id.full_id)
                )
            )
            obj.echo("\n\n")

    if dry_run:
        obj.echo("Dry-run -- not caching or aliasing")
        return

    # If this is not a dry-run, we cache the environments and write out the resolved
    # information
    update_envs = []  # type: List[ResolvedEnvironment]
    if obj.flow_datastore.TYPE != "local" or CONDA_TEST:
        # We may need to update caches
        # Note that it is possible that something we needed to resolve, we don't need
        # to cache (if we resolved to something already cached).
        formats = set()  # type: Set[str]
        for _, resolved_env, f, _ in resolver.need_caching_environments():
            update_envs.append(resolved_env)
            formats.update(f)

        conda.cache_environments(update_envs, {"conda": list(formats)})
    else:
        update_envs = [
            resolved_env for _, resolved_env, _ in resolver.new_environments()
        ]

    conda.add_environments(update_envs)

    # Update the default environment
    for env_id, resolved_env, _ in resolver.resolved_environments():
        if env_id.full_id == "_default":
            conda.set_default_environment(resolved_env.env_id)

    # We are done -- write back out the environments
    conda.write_out_environments()


@environment.command(help="Show status of environment resolution for a flow")
@click.option(
    "--local-only/--no-local-only",
    show_default=True,
    default=False,
    help="Search only locally cached environments",
)
@click.option(
    "--output",
    show_default=True,
    default=None,
    type=click.Path(file_okay=True, writable=True, resolve_path=True),
    help="If specified, the environments for the requested steps are written out "
    "to this file.",
)
@click.argument(
    "steps-to-show",
    required=False,
    nargs=-1,
)
@click.pass_obj
def show(obj: Any, local_only: bool, output: Optional[str], steps_to_show: Tuple[str]):
    # Goes from step_name to information about the step
    result = {}  # type: Dict[str, Any]
    conda = Conda(obj.echo, obj.flow_datastore.TYPE)
    for step in obj.flow:
        if not steps_to_show or step.name in steps_to_show:
            (
                _,
                step_arch,
                req,
                base_env,
            ) = CondaEnvironment.extract_merged_reqs_for_step(
                conda, obj.flow, obj.flow_datastore.TYPE, step, local_only=local_only
            )
            from_name = req.from_env_name
            result[step.name] = {
                "req": req,
                "req_hash": ResolvedEnvironment.get_req_id(
                    req.packages_as_str, req.sources
                ),
            }
            if from_name:
                result[step.name]["from_env_name"] = from_name
                if base_env is None:
                    result[step.name]["error"] = (
                        "Base environment '%s' was not found for architecture '%s'"
                        % (
                            from_name,
                            step_arch,
                        )
                    )
                    continue
                result[step.name]["base_env"] = base_env
                result[step.name]["state"] = ["derived from %s" % from_name]
            else:
                result[step.name]["state"] = []

            step_env_id = CondaEnvironment.get_env_id(conda, step.name)
            resolved_env = None
            if req.is_disabled:
                result[step.name]["state"].append("disabled")
            elif isinstance(step_env_id, EnvID):
                resolved_env = conda.environment(step_env_id, local_only)
                if resolved_env:
                    result[step.name]["env"] = resolved_env
                    result[step.name]["state"].append("resolved")

                    if (
                        obj.flow_datastore.TYPE != "local" or CONDA_TEST
                    ) and resolved_env.is_cached(
                        {
                            "conda": (
                                [CONDA_PREFERRED_FORMAT]
                                if CONDA_PREFERRED_FORMAT
                                and CONDA_PREFERRED_FORMAT != "none"
                                else ["_any"]
                            )
                        }
                    ):
                        result[step.name]["state"].append("cached")
                    if conda.created_environment(resolved_env.env_id):
                        result[step.name]["state"].append("locally present")
                else:
                    result[step.name]["state"].append("unresolved")
            elif isinstance(step_env_id, str):
                result[step.name]["state"].append("fetch-at-exec of %s" % step_env_id)
            else:
                result[step.name]["state"].append("not resolved")

    _compute_local_instances(conda)

    assert _all_local_instances is not None
    if output:
        out_json = {}
        with open(output, "w", encoding="utf-8") as f:
            for step_name, step_info in result.items():
                if "disabled" in step_info["state"]:
                    out_json[step_name] = StepEnvInfo(is_disabled=True).to_dict()
                elif "error" in step_info:
                    out_json[step_name] = StepEnvInfo(
                        is_disabled=False, error=step_info["error"]
                    ).to_dict()
                else:
                    is_fetch_at_exec = False
                    req_hash = step_info["req_hash"]
                    if any(s.startswith("fetch-at-exec") for s in step_info["state"]):
                        is_fetch_at_exec = True
                    out_json[step_name] = StepEnvInfo(
                        is_disabled=False,
                        requested_hash=req_hash,
                        requested_packages=cast(
                            StepRequirement, step_info["req"]
                        ).packages,
                        requested_sources=cast(
                            StepRequirement, step_info["req"]
                        ).sources,
                        from_env_name=step_info.get("from_env_name"),
                        is_fetch_at_exec=is_fetch_at_exec,
                        env=(
                            MetaflowResolvedEnvironment(step_info.get("env"))
                            if step_info.get("env")
                            else None
                        ),
                    ).to_dict()

            json.dump(out_json, f, indent=2)
    if not obj.is_quiet:
        for name, info in result.items():
            obj.echo("\nStep *%s*" % name)
            if "error" in info:
                obj.echo("\t%s" % info["error"], err=True)
            else:
                obj.echo("\tEnvironment is %s" % ", ".join(info["state"]))
            if "env" in info:
                env_id = cast(ResolvedEnvironment, info["env"]).env_id
                obj.echo(
                    cast(ResolvedEnvironment, info["env"]).pretty_print(
                        _all_local_instances.get(env_id.req_id, {}).get(env_id.full_id)
                    )
                )
            else:
                obj.echo(
                    "*User-requested packages* %s"
                    % ", ".join(
                        [
                            str(d).replace("*", "x")
                            for d in dict_to_tstr(
                                cast(StepRequirement, info["req"]).packages_as_str
                            )
                        ]
                    )
                )
                obj.echo(
                    "*User sources* %s"
                    % ", ".join(
                        [
                            str(s)
                            for s in dict_to_tstr(
                                cast(StepRequirement, info["req"]).sources
                            )
                        ]
                    )
                )


def _compute_local_instances(conda: Conda):
    global _all_local_instances
    if _all_local_instances is None:
        _all_local_instances = {}
        created_envs = conda.created_environments()  # This returns all MF environments
        for env_id, present_list in created_envs.items():
            _all_local_instances.setdefault(env_id.req_id, {})[env_id.full_id] = [
                os.path.basename(p) for p in present_list
            ]


# @environment.command(help="Show available environments for steps in a flow")
# @click.option(
#     "--local-only/--no-local-only",
#     show_default=True,
#     default=False,
#     help="Search only locally cached environments",
# )
# @click.option(
#     "--menu/--no-menu",
#     show_default=True,
#     default=True,
#     help="Use a menu to display the environment or print -- menu is incompatible with quiet",
# )
# @click.argument(
#     "steps-to-list",
#     required=False,
#     nargs=-1,
# )
# @click.pass_obj
# def list(
#     obj: Any,
#     local_only: bool = True,
#     menu: bool = False,
#     steps_to_list: Optional[Tuple[str]] = None,
# ):
#     conda = Conda(obj.echo, obj.flow_datastore.TYPE)
#     steps_to_potentials = _get_envs(
#         conda, obj.flow, skip_known=False, local_only=local_only, steps=steps_to_list
#     )
#     for steps_query, potentials, req_id, _ in steps_to_potentials:
#         local_instances = _local_instances_for_req_id(conda, req_id)
#         plural = "s" if len(steps_query.split(",", 1)) > 1 else ""
#         if len(potentials) == 0:
#             if not obj.is_quiet:
#                 obj.echo(
#                     "No %sknown environments for step%s %s"
#                     % (("locally " if local_only else ""), plural, steps_query)
#                 )
#             continue
#         if menu:
#             if obj.is_quiet:
#                 raise click.UsageError(
#                     message="Menu flag is incompatible with quiet flag", ctx=obj
#                 )
#             m = TerminalMenu(
#                 menu_entries=(
#                     _format_resolved_env_for_menu(conda, env)
#                     for env in potentials.values()
#                 ),
#                 title="Available environments for step%s %s (hash: %s)"
#                 % (plural, steps_query, req_id),
#                 #% _color_str(steps_query, style=_BOLD, fg=_RED),
#                 preview_title="Packages and local instances",
#                 column_headers=_columns_for_menu(),
#                 preview_cache_result=True,
#                 preview_border=True,
#                 preview_command=partial(_info_for_env, potentials, local_instances),
#                 preview_size=0.5,
#             )
#             m.show()
#         elif obj.is_quiet:
#             print(steps_query)
#             for env in potentials.values():
#                 _quiet_print_env(obj, env, local_instances.get(env.env_id.full_id))
#         else:
#             obj.echo(
#                 "### Environments for step%s *%s* (hash: %s) ###"
#                 % (plural, steps_query, req_id)
#             )
#             first_env = True
#             for env in potentials.values():
#                 if not first_env:
#                     obj.echo("\n*~~~*\n")
#                 first_env = False
#                 _pretty_print_env(
#                     obj,
#                     env,
#                     env.env_id
#                     == conda.get_default_environment(
#                         env.env_id.req_id, env.env_id.arch
#                     ),
#                     local_instances.get(env.env_id.full_id),
#                 )
#             obj.echo("\n\n")

# def _color_str(
#     s: str,
#     style: Optional[Union[int, List[int]]] = None,
#     fg: Optional[int] = None,
#     bg: Optional[int] = None,
# ) -> str:
#     if style is None:
#         style = []
#     elif isinstance(style, int):
#         style = [style]
#     if fg is not None:
#         style.append(fg)
#     if bg is not None:
#         style.append(bg + 10)
#     return "\033[%sm%s\033[0m" % (";".join((str(x) for x in style)), s)


# def _get_envs(
#     conda: Conda,
#     flow: FlowSpec,
#     skip_known: bool,
#     local_only: bool,
#     steps: Optional[Tuple[str]] = None,
# ) -> List[Tuple[str, Dict[str, ResolvedEnvironment], str, Tuple[str]]]:
#     reqs_to_steps = {}  # type: Dict[Tuple[str, Tuple[str]], List[str]]
#     for step in flow:
#         if steps and step.name not in steps:
#             continue
#         step_conda_dec = get_conda_decorator(flow, step.__name__)
#         # We want to gather steps by the type of requirements they have. This is
#         # basically their req ID as well as the requested architectures.
#         env_ids = step_conda_dec.env_ids
#         archs = tuple(sorted(env.arch for env in env_ids))
#         req_id = env_ids[0].req_id
#         reqs_to_steps.setdefault((req_id, archs), []).append(step.__name__)
#     # We now look for environments that have the proper req_id and a superset of the
#     # architectures asked for
#     if debug.conda:
#         debug.conda_exec("Requirements to resolve are:")
#         for (req_id, archs), steps in reqs_to_steps.items():
#             debug.conda_exec(
#                 "%s (archs: %s) for %s" % (req_id, ", ".join(archs), ", ".join(steps))
#             )

#     steps_to_potentials = (
#         []
#     )  # type: List[Tuple[str, Dict[str, ResolvedEnvironment], str, Tuple[str]]]
#     for (req_id, archs), steps in reqs_to_steps.items():
#         # We can pick just a single arch to look for potentials; we know we need
#         # at least that one
#         arch = archs[0]
#         need_archs = set(archs)
#         possible_envs = conda.environments(req_id, arch, local_only=local_only)
#         possible_envs = [
#             r for r in possible_envs if need_archs.issubset(r[1].co_resolved_archs)
#         ]
#         if skip_known:
#             default_env_id = conda.get_default_environment(req_id, arch)
#             if default_env_id and default_env_id in [env[0] for env in possible_envs]:
#                 debug.conda_exec(
#                     "Have known local default for %s -- skipping" % ", ".join(steps)
#                 )
#                 continue
#         steps_to_potentials.append(
#             (
#                 ", ".join(steps),
#                 {env_id.full_id: env for env_id, env in possible_envs},
#                 req_id,
#                 archs,
#             )
#         )
#     return steps_to_potentials


# def _info_for_env(
#     env_dict: Dict[str, ResolvedEnvironment],
#     local_instances: Dict[str, List[str]],
#     full_id: str,
# ) -> List[str]:
#     resolved_env = env_dict.get(full_id)
#     if resolved_env is None:
#         return [""]
#     to_return = []  # type: List[str]
#     pip_packages = []  # type: List[PackageSpecification]
#     conda_packages = []  # type: List[PackageSpecification]
#     for p in resolved_env.packages:
#         if p.TYPE == "pip":
#             pip_packages.append(p)
#         else:
#             conda_packages.append(p)

#     pip_packages.sort(key=lambda x: x.package_name)
#     conda_packages.sort(key=lambda x: x.package_name)
#     local_dirs = local_instances.get(full_id)
#     if local_dirs:
#         to_return.append(
#             _color_str("Locally Present", style=[_UNDERLINE, _BOLD], bg=_YELLOW)
#         )
#         to_return.extend(
#             (
#                 "- %s"
#                 % d.replace(resolved_env.env_id.req_id, "<req-id>").replace(
#                     full_id, "<id>"
#                 )
#                 for d in local_dirs
#             )
#         )
#     if conda_packages:
#         to_return.append(
#             _color_str("Conda Packages", style=[_UNDERLINE, _BOLD], bg=_YELLOW)
#         )
#         to_return.extend(
#             (
#                 _color_str("- %s==%s" % (p.package_name, p.package_version), fg=_GREEN)
#                 for p in conda_packages
#             )
#         )
#     if pip_packages:
#         to_return.append(
#             _color_str("Pip Packages", style=[_UNDERLINE, _BOLD], bg=_YELLOW)
#         )
#         to_return.extend(
#             (
#                 _color_str("- %s==%s" % (p.package_name, p.package_version), fg=_BLUE)
#                 for p in pip_packages
#             )
#         )
#     return to_return


# def _columns_for_menu() -> List[str]:
#     return ["Env ID", "Archs", "Creator", "Resolved On", "Tags"]


# def _unset_env_for_menu() -> List[str]:
#     return ["re-resolve", "", "", "", ""]


# def _format_resolved_env_for_menu(conda: Conda, env: ResolvedEnvironment) -> List[str]:
#     # Returns full_id, resolved_archs, resolved_by, resolved_on, tags
#     is_default = env.env_id == conda.get_default_environment(
#         env.env_id.req_id, env.env_id.arch
#     )
#     tags = env.env_alias if env.env_alias else ""
#     if is_default:
#         if tags:
#             tags += ", [default]"
#         tags = "[default]"

#     return [
#         "%s|%s" % (env.env_id.full_id, env.env_id.full_id),
#         ", ".join(env.co_resolved_archs),
#         env.resolved_by,
#         env.resolved_on.isoformat(),
#         ("[default]" if is_default else ""),
#     ]
